import os

ATASKQ_CONNECTION = os.getenv("ATASKQ_CONNECTION", "sqlite://ataskq.db.sqlite3")
ATASKQ_TASK_WAIT_TIMEOUT = os.getenv("ATASKQ_TASK_WAIT_TIMEOUT", None) and float(os.environ["ATASKQ_TASK_WAIT_TIMEOUT"])
ATASKQ_TASK_PULL_INTERVAL = float(os.getenv("ATASKQ_TASK_PULL_INTERVAL", 0.2))
ATASKQ_MONITOR_PULSE_INTERVAL = float(os.getenv("ATASKQ_MONITOR_PULSE_INTERVAL", 60))
ATASKQ_TASK_PULSE_TIMEOUT = float(os.getenv("ATASKQ_TASK_PULSE_TIMEOUT", 60 * 5))

ATASKQ_SERVER_CONNECTION = os.getenv("ATASKQ_SERVER_CONNECTION", "sqlite://ataskq.db.sqlite3")
ATASKQ_SERVER_TASK_PULSE_TIMEOUT_MONITOR_INTERVAL = float(
    os.getenv("ATASKQ_SERVER_TASK_PULSE_TIMEOUT_MONITOR_INTERVAL", 60 * 5)
)

ATASKQ_DB_LIMIT_DEFAULT = int(os.getenv("ATASKQ_DB_LIMIT_DEFAULT", 100))
ATASKQ_DB_INIT_ON_HANDLER_INIT = bool(os.getenv("ATASKQ_DB_INIT_ON_HANDLER_INIT", True))



import logging
import asyncio
from pathlib import Path

from fastapi import FastAPI, Request, Depends
from fastapi.responses import FileResponse, RedirectResponse
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles
from contextlib import asynccontextmanager


from ataskq.handler import from_connection_str
from ataskq.handler import DBHandler
from ataskq.handler.rest_handler import RESTHandler as rh
from ataskq.models import Model, __MODELS__
from ataskq.env import (
    ATASKQ_SERVER_CONNECTION,
    ATASKQ_SERVER_TASK_PULSE_TIMEOUT_MONITOR_INTERVAL,
    ATASKQ_TASK_PULSE_TIMEOUT,
)

# from .form_utils import form_data_array

logger = logging.getLogger("uvicorn")
logger.info(f"ATASKQ_SERVER_CONNECTION: {ATASKQ_SERVER_CONNECTION}")
logger.info(f"ATASKQ_TASK_PULSE_TIMEOUT: {ATASKQ_TASK_PULSE_TIMEOUT}")
logger.info(f"ATASKQ_SERVER_TASK_PULSE_TIMEOUT_MONITOR_INTERVAL: {ATASKQ_SERVER_TASK_PULSE_TIMEOUT_MONITOR_INTERVAL}")


# DB Handler


def db_handler() -> DBHandler:
    return from_connection_str(ATASKQ_SERVER_CONNECTION)


async def set_timout_tasks_task():
    dbh = db_handler()
    while True:
        logger.debug("Set Timeout Tasks")
        dbh.fail_pulse_timeout_tasks(ATASKQ_TASK_PULSE_TIMEOUT)
        await asyncio.sleep(ATASKQ_SERVER_TASK_PULSE_TIMEOUT_MONITOR_INTERVAL)


@asynccontextmanager
async def lifespan(app: FastAPI):
    logger.info("enter lifspan")

    logger.info("init db")
    db_handler().init_db()

    task = asyncio.create_task(set_timout_tasks_task())

    # Load the ML model
    yield
    # Clean up the ML models and release the resources
    logger.info("cancel task")
    task.cancel()
    logger.info("exit lifspan")


app = FastAPI(lifespan=lifespan)

# allow all cors
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

# Custom exception handler


# async def custom_exception_handler(request: Request, exc: Exception):
#     raise exc

# # Adding the custom exception handler to the app
# app.add_exception_handler(Exception, custom_exception_handler)

# Example route with intentional exception


# static folder
app.mount("/www", StaticFiles(directory=Path(__file__).parent / "www"), name="www")


@app.get("/")
async def root():
    # return "Welcome to A-TASK-Q Server"
    return RedirectResponse("/custom_query/jobs_status")


@app.get("/health")
async def health():
    return "ATASKQ Server is running"


@app.get("/api")
async def api():
    return {"message": "Welcome to A-TASK-Q Server API"}


####################
# Custom Query API #
####################
@app.get("/api/custom_query/take_next_task")
async def take_next_task(
    job_id: int,
    level_start: int = None,
    level_stop: int = None,
    dbh: DBHandler = Depends(db_handler),
):
    # take next task
    action, task = dbh.take_next_task(job_id, level_start, level_stop)
    task = rh.to_interface(task) if task is not None else None

    return dict(action=action, task=task)


@app.get("/api/custom_query/jobs_status")
async def jobs_status(request: Request, dbh: DBHandler = Depends(db_handler)):
    ret = dbh.jobs_status(**request.query_params)

    return ret


@app.get("/api/custom_query/tasks_status/{job_id}")
async def tasks_status(job_id: int, request: Request, dbh: DBHandler = Depends(db_handler)):
    ret = dbh.tasks_status(job_id=job_id, **request.query_params)

    return ret


#############
# Model API #
#############
@app.get("/api/{model}")
async def get_model_all(model: str, request: Request, dbh: DBHandler = Depends(db_handler)):
    # logger.info(f"query_params: {request.query_params}")
    model_cls = __MODELS__[model]
    mkwargs = model_cls.get_all_dict(_handler=dbh, **request.query_params)
    ikwargs = rh.m2i(model_cls, mkwargs)

    return ikwargs


@app.get("/api/{model}/count")
async def count_model_all(model: str, request: Request, dbh: DBHandler = Depends(db_handler)):
    model_cls = __MODELS__[model]
    count = model_cls.count_all(_handler=dbh, **request.query_params)

    return count


@app.get("/api/{model}/{model_id}")
async def get_model(model: str, model_id: int, dbh: DBHandler = Depends(db_handler)):
    model_cls = __MODELS__[model]
    mkwargs = model_cls.get_dict(model_id, _handler=dbh)
    ikwargs = rh.m2i(model_cls, mkwargs)

    return ikwargs


@app.post("/api/{model}")
async def create_model(model: str, request: Request, dbh: DBHandler = Depends(db_handler)):
    model_cls: Model = __MODELS__[model]
    ikwargs = await request.json()
    mkwargs = rh.i2m(model_cls, ikwargs)
    model_id = dbh.create(model_cls, **mkwargs)

    return model_id


@app.post("/api/{model}/bulk")
async def create_model_bulk(model: str, request: Request, dbh: DBHandler = Depends(db_handler)):
    model_cls: Model = __MODELS__[model]
    ikwargs = await request.json()
    mkwargs = rh.i2m(model_cls, ikwargs)
    model_ids = dbh.create_bulk(model_cls, mkwargs)

    return model_ids


@app.put("/api/{model}/{model_id}")
async def update_model(model: str, model_id: int, request: Request, dbh: DBHandler = Depends(db_handler)):
    model_cls: Model = __MODELS__[model]
    ikwargs = await request.json()
    mkwargs = rh.i2m(model_cls, ikwargs)
    dbh.update(model_cls, model_id, **mkwargs)

    return {model_cls.id_key(): model_id}


@app.delete("/api/{model}/{model_id}")
async def delete_model(model: str, model_id: int, dbh: DBHandler = Depends(db_handler)):
    model_cls: Model = __MODELS__[model]
    dbh.delete(model_cls, model_id)

    return {model_cls.id_key(): model_id}


#######
# WWW #
#######
@app.get("/db/{model}")
async def show_db(model: str):
    return FileResponse(Path(__file__).parent / "www" / "index.html")


@app.get("/custom_query/{query}")
async def show_custom_qeury(query: str):
    return FileResponse(Path(__file__).parent / "www" / "index.html")


@app.get("/custom_query/{query}/{job_id}")
async def show_job_custom_query(query: str, job_id: int):
    return FileResponse(Path(__file__).parent / "www" / "index.html")



import multiprocessing
from typing import Union
import pickle
import logging
from importlib import import_module
from multiprocessing import Process
from inspect import signature
import time
from typing import Dict, List
from datetime import datetime

from .env import (
    ATASKQ_CONNECTION,
    ATASKQ_MONITOR_PULSE_INTERVAL,
    ATASKQ_TASK_PULSE_TIMEOUT,
    ATASKQ_TASK_PULL_INTERVAL,
    ATASKQ_TASK_WAIT_TIMEOUT,
)
from .logger import Logger
from .models import EStatus, Job, StateKWArg, Task, EntryPointRuntimeError
from .monitor import MonitorThread
from .handler import Handler, DBHandler, from_connection_str, EAction


def targs(*args, **kwargs):
    return (args, kwargs)


class TaskQ(Logger):
    def __init__(
        self,
        job_id=None,
        conn=ATASKQ_CONNECTION,
        run_task_raise_exception=False,
        task_wait_timeout=ATASKQ_TASK_WAIT_TIMEOUT,
        task_pull_intervnal=ATASKQ_TASK_PULL_INTERVAL,
        monitor_pulse_interval=ATASKQ_MONITOR_PULSE_INTERVAL,
        task_pulse_timeout=ATASKQ_TASK_PULSE_TIMEOUT,
        max_jobs: int = None,
        logger: Union[logging.Logger, None] = None,
    ) -> None:
        """
        Args:
        task_pull_intervnal: pulling interval for task to complete in seconds.
        task_pulse_timeout: update interval for pulse in seconds while taks is running.
        monitor_timeout_internal: timeout for task last monitor pulse in seconds, if passed before getting next task, set to Failure.
        run_task_raise_exception: if True, run_task will raise exception when task fails. This is for debugging purpose only and will fail production flow.
        """
        super().__init__(logger)

        # init db handler
        # todo: hanlder shouldn't get job_id
        self._handler: Handler = (
            conn if isinstance(conn, Handler) else from_connection_str(conn=conn, logger=self._logger)
        )

        # get job
        if job_id is not None:
            job = Job.get(self._job_id, self._handler)
        else:
            job = None
        self._job = job

        self._run_task_raise_exception = run_task_raise_exception
        self._task_wait_timeout = task_wait_timeout
        self._task_pull_interval = task_pull_intervnal
        self._monitor_pulse_interval = monitor_pulse_interval
        self._task_pulse_timeout = task_pulse_timeout
        self._max_jobs = max_jobs

        # state kwargs for jobs
        self._state_kwargs: Dict[str:object] = dict()

        self._running = False

    @property
    def job(self):
        return self._job

    @property
    def job_id(self):
        return self._job.job_id

    @property
    def handler(self):
        return self._handler

    @property
    def task_wait_interval(self):
        return self._task_pull_interval

    @property
    def monitor_pulse_interval(self):
        return self._monitor_pulse_interval

    def clear_job(self):
        self._job = None

    def set_job(self, job_id):
        job = Job.get(job_id, self._handler)
        self._handler.set_job_id(job.job_id)  # todo remove - handler should need job id ...
        self._job = job

    def create_job(self, name=None, description=None):
        assert self._job is None, "job already assigned to current taskq, run clear_job to create new job."

        job = Job(name=name, description=description).create(_handler=self._handler)
        self._job = job

        if self._max_jobs is not None:
            # keep max jbos
            Job.delete_all(
                _where=f"job_id NOT IN (SELECT job_id FROM jobs ORDER BY job_id DESC limit {self._max_jobs})",
                _handler=self.handler,
            )

        return self

    def delete_job(self):
        self.job.delete(_handler=self.handler)
        self._job = None

    def get_state_kwargs(self) -> List[StateKWArg]:
        return self.job.get_state_kwargs(self._handler)

    def add_state_kwargs(self, state_kwargs):
        self.job.add_state_kwargs(state_kwargs, _handler=self._handler)

        return self

    def get_tasks(self) -> List[Task]:
        return self.job.get_tasks(self._handler)

    def add_tasks(self, tasks):
        self.job.add_tasks(tasks, _handler=self._handler)

        return self

    def update_task_start_time(self, task: Task, start_time: datetime = None):
        if start_time is None:
            start_time = datetime.now()

        task.update(start_time=start_time, _handler=self._handler)

    def update_task_status(self, task: Task, status: EStatus, timestamp: datetime = None):
        if timestamp is None:
            timestamp = datetime.now()

        if status == EStatus.RUNNING:
            # for running task update pulse_time
            task.update(_handler=self._handler, status=status, pulse_time=timestamp)
        elif status == EStatus.SUCCESS or status == EStatus.FAILURE:
            # for done task update pulse_time and done_time time as well
            task.update(_handler=self._handler, status=status, pulse_time=timestamp, done_time=timestamp)
        else:
            raise RuntimeError(f"Unsupported status '{status}' for status update")

    def count_pending_tasks_below_level(self, level):
        ret = Task.count_all(
            _where=f"job_id = {self.job_id} AND level < {level} AND status in ('{EStatus.PENDING}')",
            _handler=self._handler,
        )
        return ret

    def _run_task(self, task: Task):
        # get entry point func to execute
        ep = task.entrypoint
        if ep == "ataskq.skip_run_task":
            self.info(f"task '{task.task_id}' is marked as 'skip_run_task', skipping run task.")
            return

        assert "." in ep, "entry point must be inside a module."
        module_name, func_name = ep.rsplit(".", 1)
        try:
            m = import_module(module_name)
        except ImportError as ex:
            raise RuntimeError(f"Failed to load module '{module_name}'. Exception: '{ex}'")
        assert hasattr(
            m, func_name
        ), f"failed to load entry point, module '{module_name}' doen't have func named '{func_name}'."
        func = getattr(m, func_name)
        assert callable(func), f"entry point is not callable, '{module_name},{func}'."

        # get targs
        if task.targs is not None:
            try:
                targs = pickle.loads(task.targs)
            except Exception as ex:
                if self._run_task_raise_exception:  # for debug purposes only
                    self.warning("Getting tasks args failed.")
                    self.update_task_status(task, EStatus.FAILURE)
                    raise ex

                self.warning("Getting tasks args failed.", exc_info=True)
                self.update_task_status(task, EStatus.FAILURE)

                return
        else:
            targs = ((), {})

        # get state kwargs for kwarg in function signature
        ep_state_kwargs = {
            param.name: param.default
            for param in signature(func).parameters.values()
            if param.default is not param.empty and param.name in self._state_kwargs
        }

        for name, default in ep_state_kwargs.items():
            state_kwarg = self._state_kwargs.get(name)
            if isinstance(state_kwarg, StateKWArg):
                # state kwargs not initalized yet
                assert default is None, f"state kwarg '{name}' default value must be None"
                self._logger.info(f"Initializing state kwarg '{name}'")
                try:
                    self._state_kwargs[name] = state_kwarg.call()
                except EntryPointRuntimeError as ex:
                    self._logger.info(f"Failed to initialize state kwarg '{name}'")
                    self.update_task_status(task, EStatus.FAILURE)

                    if self._run_task_raise_exception:
                        raise ex
                    return
            # state kwargs already inistliazed
            self._logger.info(f"Update kwarg '{name}' with state kwargs")
            targs[1][name] = self._state_kwargs[name]

        # update task start time
        self.update_task_start_time(task)

        # run task
        monitor = MonitorThread(task, self, pulse_interval=self._monitor_pulse_interval)
        monitor.start()

        try:
            func(*targs[0], **targs[1])
            status = EStatus.SUCCESS
        except Exception as ex:
            if self._run_task_raise_exception:  # for debug purposes only
                self.warning("Running task entry point failed with exception.")
                self.update_task_status(task, EStatus.FAILURE)
                monitor.stop()
                monitor.join()
                raise ex

            self.warning("Running task entry point failed with exception.", exc_info=True)
            status = EStatus.FAILURE

        monitor.stop()
        monitor.join()
        self.update_task_status(task, status)

    @staticmethod
    def init_state_kwarg(self, state_kwarg: StateKWArg):
        return

    def _take_next_task(self, level):
        assert self._job is not None, "job must be assigned to taskq before taking next task."
        level_start = level.start if level is not None else None
        level_stop = level.stop if level is not None else None

        return self._handler.take_next_task(self.job_id, level_start, level_stop)

    def _run(self, level):
        # make sure all state kwargs for job have key in self._state_kwargs, later to be used in _run_task
        state_kwargs_db = self.job.get_state_kwargs(self._handler)
        for skw in state_kwargs_db:
            if skw.name not in self._state_kwargs:
                self._state_kwargs[skw.name] = skw

        # check for error code
        task_pull_start = time.time()
        while True:
            # if the taskq handler is db handler, the taskq performs background tasks before each run
            if isinstance(self._handler, DBHandler):
                self._handler.fail_pulse_timeout_tasks(self._task_pulse_timeout)
            # grab tasks and set them in Q
            action, task = self._take_next_task(level)

            # handle no task available
            if action == EAction.STOP:
                break
            if action == EAction.RUN_TASK:
                self._run_task(task)
            elif action == EAction.WAIT:
                if self._task_wait_timeout is not None and time.time() - task_pull_start > self._task_wait_timeout:
                    raise Exception(f"task pull timeout of '{self._task_wait_timeout}' sec reached.")

                self.debug(f"waiting for {self._task_pull_interval} sec before taking next task")
                time.sleep(self._task_pull_interval)

    def assert_level(self, level):
        if isinstance(level, int):
            level = range(level, level + 1)
        elif isinstance(level, (list, tuple)):
            assert len(level) == 2, "level of type list or tuple must have length of 2"
            level = range(level[0], level[1])
        else:
            assert isinstance(level, range), "level must be int, list, tuple or range"

        # check all task < level.start are done
        count = self.count_pending_tasks_below_level(level.start)
        assert count == 0, f"all tasks below level must be done before running tasks at levels {level}"

        return level

    def run(self, num_processes=None, level=None):
        if level is not None:
            level = self.assert_level(level)

        self._running = True

        # default to run in current process
        if num_processes is None:
            self._run(level)
            self._running = False
            return

        assert isinstance(num_processes, (int, float))

        if isinstance(num_processes, float):
            assert 0.0 <= num_processes <= 1.0
            nprocesses = int(multiprocessing.cpu_count() * num_processes)
        elif num_processes < 0:
            nprocesses = multiprocessing.cpu_count() - num_processes
        else:
            nprocesses = num_processes

        # set processes and Q
        processes = [Process(target=self._run, args=(level,)) for i in range(nprocesses)]
        [p.start() for p in processes]

        # join all processes
        [p.join() for p in processes]

        # log failed processes
        fail = False
        for p in processes:
            if p.exitcode != 0:
                fail = True
                self.error(f"Process '{p.pid}' failed with exitcode '{p.exitcode}'")

        self._running = False

        if self._run_task_raise_exception and fail:
            raise Exception("Some processes failed, see logs for details")

        return self



import re
from typing import NamedTuple
import sqlite3
from datetime import datetime

from ..models import Model
from .handler import to_datetime, from_datetime
from .db_handler import DBHandler


class SqliteConnection(NamedTuple):
    path: str

    def __str__(self):
        return f"sqlite://{self.path}"


def from_connection_str(conn):
    format = "sqlite://path"
    pattern = r"sqlite://(?P<path>.+)$"
    match = re.match(pattern, conn)

    if not match:
        raise Exception(f"conn must be in '{format}', ex: 'sqlite://ataskq.db.sqlite3'")

    path = match.group("path")
    ret = SqliteConnection(path=path)

    return ret


class SQLite3DBHandler(DBHandler):
    def __init__(self, conn=None, **kwargs) -> None:
        self._connection = from_connection_str(conn)
        super().__init__(**kwargs)

    @staticmethod
    def m2i_serialize():
        type_handlers = {
            datetime: lambda v: from_datetime(v),
        }

        return type_handlers

    @staticmethod
    def i2m_serialize():
        type_handlers = {
            datetime: lambda v: to_datetime(v),
        }

        return type_handlers

    @property
    def pragma_foreign_keys_on(self):
        return "PRAGMA foreign_keys = ON"

    @property
    def format_symbol(self):
        return "?"

    @property
    def connection(self):
        return self._connection

    @property
    def db_path(self):
        return self._connection.path

    @property
    def bytes_type(self):
        return "MEDIUMBLOB"

    @property
    def primary_key(self):
        return "INTEGER PRIMARY KEY AUTOINCREMENT"

    @property
    def timestamp_type(self):
        return "DATETIME"

    def timestamp(self, ts):
        return f"'{ts}'"

    @property
    def begin_exclusive(self):
        return "BEGIN EXCLUSIVE"

    @property
    def for_update(self):
        return ""

    def connect(self):
        conn = sqlite3.connect(self.db_path)
        return conn



from pathlib import Path
from datetime import datetime, timedelta
from copy import copy

import pytest

from ataskq import TaskQ, StateKWArg, Job, Task, targs, EStatus
from ataskq.handler import DBHandler
from ataskq.handler import EAction, from_connection_str

from .tasks_utils import dummy_args_task


@pytest.fixture
def jtaskq(conn) -> TaskQ:
    return TaskQ(conn=conn).create_job()


def test_create_job(conn):
    taskq = TaskQ(conn=conn).create_job()
    assert isinstance(taskq, TaskQ)

    if "sqlite" in conn:
        assert Path(taskq.handler.db_path).exists()
        assert Path(taskq.handler.db_path).is_file()
    elif "pg" in conn:
        pass
    elif "http" in conn:
        pass
    else:
        raise Exception(f"unknown db type in connection string '{conn}'")


def test_job_default_name(conn):
    job = TaskQ(conn=conn).create_job().job
    assert job.name is None


def test_job_custom_name(conn):
    job = TaskQ(conn=conn).create_job(name="my_job").job
    assert job.name == "my_job"


def test_task_job_delete_cascade(conn):
    # test that deleting a job deletes all its tasks
    handler = from_connection_str(conn=conn)
    taskq1: TaskQ = TaskQ(conn=handler).create_job(name="job1")
    taskq2: TaskQ = TaskQ(conn=handler).create_job(name="job2")
    assert Job.count_all(_handler=handler) == 2

    taskq1.add_tasks(
        [
            Task(entrypoint=""),
            Task(entrypoint=""),
            Task(entrypoint=""),
        ]
    )
    assert Task.count_all(_handler=handler) == 3
    assert len(taskq1.get_tasks()) == 3

    taskq2.add_tasks(
        [
            Task(entrypoint=""),
            Task(entrypoint=""),
        ]
    )
    assert Task.count_all(_handler=handler) == 5
    assert len(taskq1.get_tasks()) == 3
    assert len(taskq2.get_tasks()) == 2

    taskq2.delete_job()
    assert len(Job.get_all(_handler=handler)) == 1

    assert Task.count_all(_handler=handler) == 3
    assert len(taskq1.get_tasks()) == 3

    taskq1.delete_job()
    assert Task.count_all(_handler=handler) == 0


def test_state_kwargs_job_delete_cascade(conn):
    # todo: test should ne under ataskq
    handler = from_connection_str(conn=conn)
    taskq1: TaskQ = TaskQ(conn=handler).create_job(name="job1")
    taskq2: TaskQ = TaskQ(conn=handler).create_job(name="job2")
    assert len(Job.get_all(_handler=handler)) == 2

    taskq1.add_state_kwargs(
        [
            StateKWArg(entrypoint=""),
            StateKWArg(entrypoint=""),
            StateKWArg(entrypoint=""),
        ]
    )
    assert len(StateKWArg.get_all(_handler=handler)) == 3
    assert len(taskq1.get_state_kwargs()) == 3

    taskq2.add_state_kwargs(
        [
            StateKWArg(entrypoint=""),
            StateKWArg(entrypoint=""),
        ]
    )
    assert len(StateKWArg.get_all(_handler=handler)) == 5
    assert len(taskq1.get_state_kwargs()) == 3
    assert len(taskq2.get_state_kwargs()) == 2

    taskq2.delete_job()
    assert len(Job.get_all(_handler=handler)) == 1

    assert len(StateKWArg.get_all(_handler=handler)) == 3
    assert len(taskq1.get_state_kwargs()) == 3

    taskq1.delete_job()
    assert len(StateKWArg.get_all(_handler=handler)) == 0


def test_update_task_start_time(jtaskq):
    in_task1 = Task(entrypoint="ataskq.tasks_utils.dummy_args_task", level=1, name="task1")

    jtaskq.add_tasks(
        [
            in_task1,
        ]
    )

    tasks = jtaskq.get_tasks()
    assert len(tasks) == 1  # sanity
    task: Task = tasks[0]

    # update db with task time (also updates task inplace)
    start_time = datetime.now()
    start_time = start_time.replace(microsecond=0)  # test to seconds resolution
    jtaskq.update_task_start_time(task, start_time)
    assert task.start_time == start_time

    # get task form db and validate
    tasks = jtaskq.get_tasks()
    assert len(tasks) == 1  # sanity
    task: Task = tasks[0]
    assert task.start_time == start_time


def test_update_task_status(jtaskq):
    in_task1 = Task(entrypoint="ataskq.tasks_utils.dummy_args_task", level=1, name="task1")

    jtaskq.add_tasks(
        [
            in_task1,
        ]
    )

    tasks = jtaskq.get_tasks()
    assert len(tasks) == 1  # sanity
    task: Task = tasks[0]
    assert task.status == EStatus.PENDING

    now = datetime.now()
    # update db with task status (also updates task inplace)
    jtaskq.update_task_status(task, EStatus.RUNNING, timestamp=now)
    assert task.status == EStatus.RUNNING
    assert task.pulse_time == now
    assert task.done_time is None
    # get task form db and validate
    tasks = jtaskq.get_tasks()
    assert len(tasks) == 1  # sanity
    task: Task = tasks[0]
    assert task.status == EStatus.RUNNING
    assert task.pulse_time == now
    assert task.done_time is None

    # update db with task status (also updates task inplace)
    jtaskq.update_task_status(task, EStatus.SUCCESS, timestamp=now)
    assert task.status == EStatus.SUCCESS
    assert task.pulse_time == now
    assert task.done_time == now
    # get task form db and validate
    tasks = jtaskq.get_tasks()
    assert len(tasks) == 1  # sanity
    task: Task = tasks[0]
    assert task.status == EStatus.SUCCESS
    assert task.pulse_time == now
    assert task.done_time == now


def _compare_task_taken(task1: Task, task2: Task, job_id=None):
    if job_id:
        assert task1.job_id == job_id
    assert task1.job_id == task2.job_id
    assert task1.name == task2.name
    assert task1.description == task2.description
    assert task1.level == task2.level
    assert task1.entrypoint == task2.entrypoint
    assert task1.targs == task2.targs


def test_take_next_task_sanity(jtaskq):
    in_task1 = Task(entrypoint="ataskq.tasks_utils.dummy_args_task", level=1, name="task1")

    jtaskq.add_tasks(
        [
            in_task1,
        ]
    )

    action, task = jtaskq._take_next_task(level=None)
    assert action == EAction.RUN_TASK
    _compare_task_taken(in_task1, task)


def test_take_next_task(jtaskq):
    in_task1 = Task(entrypoint="ataskq.tasks_utils.dummy_args_task", level=1, name="task1")
    in_task2 = Task(entrypoint="ataskq.tasks_utils.dummy_args_task", level=2, name="task2")
    in_task3 = copy(in_task1)

    jtaskq.add_tasks(
        [
            in_task2,
            in_task1,
            in_task3,
        ]
    )

    tids = []

    action, task = jtaskq._take_next_task(level=None)
    assert action == EAction.RUN_TASK
    assert task.task_id not in tids
    tids.append(task.task_id)
    _compare_task_taken(in_task1, task)

    action, task = jtaskq._take_next_task(level=None)
    assert action == EAction.RUN_TASK
    assert task.task_id not in tids
    tids.append(task.task_id)
    _compare_task_taken(in_task3, task)  # note in_task3 is copy of in_task1

    action, task = jtaskq._take_next_task(level=None)
    assert action == EAction.WAIT
    assert task is None


def test_take_next_task_2_jobs(conn):
    # todo: test should ne under ataskq
    handler = from_connection_str(conn=conn)
    taskq1: TaskQ = TaskQ(conn=handler).create_job(name="job1")
    taskq2: TaskQ = TaskQ(conn=handler).create_job(name="job2")

    in_task1 = Task(entrypoint="ataskq.tasks_utils.dummy_args_task", level=2, name="taska")
    in_task2 = Task(entrypoint="ataskq.tasks_utils.dummy_args_task", level=1, name="taskb")
    in_task3 = Task(entrypoint="ataskq.tasks_utils.dummy_args_task", level=1, name="taskb")
    in_task4 = Task(entrypoint="ataskq.tasks_utils.dummy_args_task", level=1, name="taskd")
    in_task5 = Task(entrypoint="ataskq.tasks_utils.dummy_args_task", level=2, name="taske")

    taskq1.add_tasks(
        [
            in_task2,
            in_task1,
            in_task3,
        ]
    )

    taskq2.add_tasks(
        [
            in_task5,
            in_task4,
        ]
    )

    jid1 = taskq1.job_id
    jid2 = taskq2.job_id
    tids = []

    # sanity check
    assert len(Job.get_all(_handler=handler)) == 2
    assert len(taskq1.get_tasks()) == 3
    assert len(Job.get_all(_handler=handler)) == 2
    assert len(taskq2.get_tasks()) == 2

    # db handler 1
    action, task = taskq1._take_next_task(level=None)
    assert action == EAction.RUN_TASK
    assert task.task_id not in tids
    tids.append(task.task_id)
    _compare_task_taken(in_task2, task, job_id=jid1)
    tids.append(task.task_id)

    action, task = taskq1._take_next_task(level=None)
    assert action == EAction.RUN_TASK
    assert task.task_id not in tids
    tids.append(task.task_id)
    _compare_task_taken(in_task3, task, job_id=jid1)  # note in_task3 is copy of in_task1

    action, task = taskq1._take_next_task(level=None)
    assert action == EAction.WAIT
    assert task is None

    # db handler 2
    action, task = taskq2._take_next_task(level=None)
    assert action == EAction.RUN_TASK
    assert task.task_id not in tids
    tids.append(task.task_id)
    _compare_task_taken(in_task4, task, job_id=jid2)
    tids.append(task.task_id)

    action, task = taskq2._take_next_task(level=None)
    assert action == EAction.WAIT
    assert task is None


def test_get_tasks(jtaskq):
    in_task1 = Task(entrypoint=dummy_args_task, level=1, name="task1")
    in_task2 = Task(entrypoint=dummy_args_task, level=2, name="task2")
    in_task3 = Task(entrypoint=dummy_args_task, level=3, name="task3")

    jtaskq.add_tasks(
        [
            in_task3,
            in_task2,
            in_task1,
        ]
    )

    tasks = jtaskq.get_tasks()
    assert len(tasks) == 3
    for t in tasks:
        if t.level == 1:
            _compare_task_taken(in_task1, t)
        elif t.level == 2:
            _compare_task_taken(in_task2, t)
        elif t.level == 3:
            _compare_task_taken(in_task3, t)


def test_run_default(conn, tmp_path: Path):
    filepath = tmp_path / "file.txt"

    taskq = TaskQ(conn=conn).create_job()

    taskq.add_tasks(
        [
            Task(entrypoint="ataskq.tasks_utils.write_to_file_tasks.write_to_file", targs=targs(filepath, "task 0\n")),
            Task(entrypoint="ataskq.tasks_utils.write_to_file_tasks.write_to_file", targs=targs(filepath, "task 1\n")),
            Task(entrypoint="ataskq.tasks_utils.write_to_file_tasks.write_to_file", targs=targs(filepath, "task 2\n")),
        ]
    )

    taskq.run()

    assert filepath.exists()
    assert filepath.read_text() == "task 0\n" "task 1\n" "task 2\n"


def test_run_task_raise_exception(conn):
    # no exception raised
    try:
        taskq: TaskQ = TaskQ(conn=conn, run_task_raise_exception=False).create_job()
        taskq.add_tasks(
            [
                Task(entrypoint="ataskq.tasks_utils.exception_task", targs=targs(message="task failed")),
            ]
        )
        taskq.run()
    except Exception:
        assert False, "exception_task raises exception with run_task_raise_exception=False"

    # exception raised
    taskq: TaskQ = TaskQ(conn=conn, run_task_raise_exception=True).create_job()
    taskq.add_tasks(
        [
            Task(entrypoint="ataskq.tasks_utils.exception_task", targs=targs(message="task failed")),
        ]
    )
    with pytest.raises(Exception) as excinfo:
        taskq.run()
    assert excinfo.value.args[0] == "task failed"


def test_run_2_processes(conn, tmp_path: Path):
    filepath = tmp_path / "file.txt"

    taskq = TaskQ(conn=conn).create_job()

    taskq.add_tasks(
        [
            Task(
                entrypoint="ataskq.tasks_utils.write_to_file_tasks.write_to_file_mp_lock",
                targs=targs(filepath, "task 0\n"),
            ),
            Task(
                entrypoint="ataskq.tasks_utils.write_to_file_tasks.write_to_file_mp_lock",
                targs=targs(filepath, "task 1\n"),
            ),
            Task(
                entrypoint="ataskq.tasks_utils.write_to_file_tasks.write_to_file_mp_lock",
                targs=targs(filepath, "task 2\n"),
            ),
        ]
    )

    taskq.run(num_processes=2)

    assert filepath.exists()
    text = filepath.read_text()
    assert "task 0\n" in text
    assert "task 1\n" in text
    assert "task 1\n" in text


@pytest.mark.parametrize("num_processes", [None, 2])
def test_run_by_level(conn, tmp_path: Path, num_processes: int):
    filepath = tmp_path / "file.txt"

    taskq = TaskQ(conn=conn).create_job()

    taskq.add_tasks(
        [
            Task(
                level=0,
                entrypoint="ataskq.tasks_utils.write_to_file_tasks.write_to_file_mp_lock",
                targs=targs(filepath, "task 0\n"),
            ),
            Task(
                level=1,
                entrypoint="ataskq.tasks_utils.write_to_file_tasks.write_to_file_mp_lock",
                targs=targs(filepath, "task 1\n"),
            ),
            Task(
                level=1,
                entrypoint="ataskq.tasks_utils.write_to_file_tasks.write_to_file_mp_lock",
                targs=targs(filepath, "task 2\n"),
            ),
            Task(
                level=2,
                entrypoint="ataskq.tasks_utils.write_to_file_tasks.write_to_file_mp_lock",
                targs=targs(filepath, "task 3\n"),
            ),
        ]
    )

    assert taskq.count_pending_tasks_below_level(3) == 4

    assert taskq.count_pending_tasks_below_level(1) == 1
    taskq.run(level=0, num_processes=num_processes)
    taskq.count_pending_tasks_below_level(1) == 0
    assert filepath.exists()
    text = filepath.read_text()
    assert "task 0\n" in text

    assert taskq.count_pending_tasks_below_level(2) == 2
    taskq.run(level=1, num_processes=num_processes)
    taskq.count_pending_tasks_below_level(2) == 0
    text = filepath.read_text()
    assert "task 0\n" in text
    assert "task 1\n" in text
    assert "task 2\n" in text

    assert taskq.count_pending_tasks_below_level(3) == 1
    taskq.run(level=2, num_processes=num_processes)
    taskq.count_pending_tasks_below_level(3) == 0
    text = filepath.read_text()
    assert "task 0\n" in text
    assert "task 1\n" in text
    assert "task 2\n" in text
    assert "task 3\n" in text


def test_monitor_pulse_failure(conn):
    # set monitor pulse longer than timeout
    taskq = TaskQ(conn=conn, monitor_pulse_interval=10, task_pulse_timeout=1.5).create_job()
    taskq.add_tasks(
        [
            # reserved keyward for ignored task for testing
            Task(entrypoint="ataskq.skip_run_task", targs=targs("task will fail")),
            Task(entrypoint="ataskq.tasks_utils.dummy_args_task", targs=targs("task will success")),
        ]
    )
    start = datetime.now()
    taskq.run()
    stop = datetime.now()

    tasks = taskq.get_tasks()

    assert tasks[0].status == EStatus.FAILURE
    assert tasks[1].status == EStatus.SUCCESS
    assert stop - start > timedelta(seconds=1.5)


def test_task_wait_timeout(conn):
    # set monitor pulse longer than timeout
    taskq = TaskQ(conn=conn, run_task_raise_exception=True, task_wait_timeout=0).create_job()
    taskq.add_tasks(
        [
            Task(entrypoint="ataskq.tasks_utils.dummy_args_task", level=1, targs=targs("task will success", sleep=0.2)),
            Task(entrypoint="ataskq.tasks_utils.dummy_args_task", level=2, targs=targs("task will success")),
        ]
    )

    with pytest.raises(Exception) as excinfo:
        taskq.run(num_processes=2)
    assert excinfo.value.args[0] == "Some processes failed, see logs for details"


def test_run_with_state_kwargs(conn):
    from ataskq.tasks_utils.counter_task import counter_kwarg, counter_task

    taskq = TaskQ(conn=conn, run_task_raise_exception=True).create_job()

    taskq.add_state_kwargs(
        [
            StateKWArg(entrypoint=counter_kwarg, name="counter"),
        ]
    )

    taskq.add_tasks(
        [
            Task(entrypoint=counter_task, targs=targs(print_counter=True)),
        ]
    )

    taskq.run()


def test_max_jobs(conn):
    max_jobs = 10
    taskq = TaskQ(conn=conn, max_jobs=max_jobs)
    if not isinstance(taskq.handler, DBHandler):
        pytest.skip()

    jobs_id = []
    for i in range(max_jobs * 2):
        taskq.clear_job()
        taskq.create_job(name=f"job{i}")
        jobs_id.append(taskq.job.job_id)
    jobs = Job.get_all(taskq.handler)
    assert len(jobs) == 10

    remaining_jobs = [j.job_id for j in jobs]
    assert remaining_jobs == jobs_id[-max_jobs:]



from pytest import fixture
import os
import pytest

if os.getenv("_PYTEST_RAISE", "0") != "0":

    @pytest.hookimpl(tryfirst=True)
    def pytest_exception_interact(call):
        raise call.excinfo.value

    @pytest.hookimpl(tryfirst=True)
    def pytest_internalerror(excinfo):
        raise excinfo.value


def truncate_query(table_name):
    return f"""DO $$
BEGIN
   IF EXISTS (SELECT 1 FROM information_schema.tables WHERE table_schema = 'public' AND table_name = '{table_name}') THEN
      TRUNCATE TABLE {table_name} CASCADE;
   END IF;
END $$;
"""


def drop_pg_tables(conn):
    assert "pg" in conn
    import psycopg2
    from ataskq.handler.postgresql import from_connection_str

    connection = from_connection_str(conn)
    db_conn = psycopg2.connect(
        host=connection.host, database=connection.database, user=connection.user, password=connection.password
    )
    c = db_conn.cursor()
    c.execute(truncate_query("tasks"))
    c.execute(truncate_query("state_kwargs"))
    c.execute(truncate_query("jobs"))
    # c.execute('DROP TABLE IF EXISTS tasks;')
    # c.execute('DROP TABLE IF EXISTS state_kwargs;')
    # c.execute('DROP TABLE IF EXISTS jobs;')
    db_conn.commit()
    db_conn.close()


@fixture
def conn(tmp_path):
    conn = os.getenv("ATASKQ_CONNECTION", "sqlite://{tmp_path}/ataskq.db.sqlite3")
    if "sqlite" in conn:
        conn = conn.format(tmp_path=tmp_path)
    elif "pg" in conn:
        # connect and clear all db tables
        drop_pg_tables(conn)
    elif "http" in conn:
        server_conn = os.getenv("ATASKQ_SERVER_CONNECTION", "pg://postgres:postgres@localhost/postgres")
        assert "pg" in server_conn, "rest api test must be with posgres server"
        drop_pg_tables(server_conn)
    else:
        raise Exception(f"Unkown connection format '{conn}'")

    return conn



from datetime import datetime, timedelta
from typing import List, Tuple
from pathlib import Path
from io import TextIOWrapper
from abc import abstractmethod
from datetime import datetime

from .handler import Handler, get_query_kwargs
from ..imodel import IModel
from ..env import ATASKQ_DB_INIT_ON_HANDLER_INIT, ATASKQ_DB_LIMIT_DEFAULT
from .. import __schema_version__

CONNECTION_LOG = [None]


def set_connection_log(log):
    CONNECTION_LOG[0] = log


def transaction_decorator(exclusive=False):
    def decorator(func):
        def wrapper(self, *args, **kwargs):
            with self.connect() as conn:
                if CONNECTION_LOG[0] is not None:
                    conn.set_trace_callback(CONNECTION_LOG[0])

                c = conn.cursor()
                try:
                    if self.pragma_foreign_keys_on:
                        c.execute(self.pragma_foreign_keys_on)

                    if exclusive:
                        c.execute(self.begin_exclusive)
                    else:
                        c.execute("BEGIN")
                    # enable foreign keys for each connection (sqlite default is off)
                    # https://www.sqlite.org/foreignkeys.html
                    # Foreign key constraints are disabled by default (for backwards
                    # compatibility), so must be enabled separately for each database
                    # connection
                    ret = func(self, c, *args, **kwargs)
                    conn.commit()
                except Exception as e:
                    self.error(f"Failed to execute transaction '{type(e)}:{e}'. Rolling back")
                    conn.rollback()
                    raise e

            return ret

        return wrapper

    return decorator


def _field_with_order(f):
    if isinstance(f, (tuple, list)):
        if len(f) == 1:
            return f"{f[0]} ASC"
        assert len(f) == 2, "order_by tuple must be of format (field, order) where order default is ASC"

    else:
        return f"{f} ASC"
    return f"{f[0]} {f[1]}"


def order_query(order_by):
    if isinstance(order_by, str):
        return order_by

    assert isinstance(order_by, (tuple, list)), "order_by must be of type str, tuple or list"

    order_by = [_field_with_order(f) for f in order_by]
    order_by = ", ".join(order_by)

    return order_by


def expand_query_str(query_str, _where=None, _order_by=None, _limit=None, _offset=None):
    if _where is not None:
        query_str += f" WHERE {_where}"

    if _order_by is not None:
        query_str += f" ORDER BY {order_query(_order_by)}"

    if _limit is not None:
        query_str += f" LIMIT {_limit}"

    if _offset is not None:
        query_str += f" OFFSET {_offset}"

    return query_str


class DBHandler(Handler):
    def __init__(self, init_db=ATASKQ_DB_INIT_ON_HANDLER_INIT, logger=None) -> None:
        super().__init__(logger)

        if init_db:
            self.init_db()

    @property
    def db_path(self):
        raise Exception(f"'{self.__class__.__name__}' db doesn't support db path property'")

    def count_all(self, model_cls: IModel, **kwargs):
        query_kwargs = get_query_kwargs(kwargs)
        count = self.count_query(model_cls, **query_kwargs)

        return count

    def get_all(self, model_cls: IModel, **kwargs) -> List[dict]:
        query_kwargs = get_query_kwargs(kwargs)
        rows, col_names, _ = self.select_query(model_cls, **query_kwargs)
        ret = [dict(zip(col_names, row)) for row in rows]

        return ret

    def get(self, model_cls: IModel, model_id) -> dict:
        rows, col_names, query_str = self.select_query(model_cls, _where=f"{model_cls.id_key()} = {model_id}")
        assert len(rows) != 0, f"no match found for '{model_cls.__name__}', query: '{query_str}'."
        assert len(rows) == 1, f"more than 1 row found for '{model_cls.__name__}', query: '{query_str}'."
        ret = [dict(zip(col_names, row)) for row in rows][0]

        return ret

    @property
    def pragma_foreign_keys_on(self):
        return None

    @property
    @abstractmethod
    def format_symbol(self):
        pass

    @property
    @abstractmethod
    def connection(self):
        pass

    @property
    @abstractmethod
    def bytes_type(self):
        pass

    @property
    @abstractmethod
    def primary_key(self):
        pass

    @property
    @abstractmethod
    def timestamp_type(self):
        pass

    @abstractmethod
    def timestamp(self, ts):
        pass

    @property
    @abstractmethod
    def begin_exclusive(self):
        pass

    @property
    @abstractmethod
    def for_update(self):
        pass

    @abstractmethod
    def connect(self):
        pass

    @transaction_decorator()
    def _create(self, c, model_cls: IModel, **ikwargs) -> int:
        d = {k: v for k, v in ikwargs.items() if model_cls.id_key() not in k}
        keys = list(d.keys())
        values = list(d.values())
        c.execute(
            f'INSERT INTO {model_cls.table_key()} ({", ".join(keys)}) VALUES ({", ".join([self.format_symbol] * len(keys))}) RETURNING {model_cls.id_key()}',
            values,
        )

        model_id = c.fetchone()[0]

        return model_id

    @transaction_decorator()
    def _create_bulk(self, c, model_cls: IModel, ikwargs: List[dict]) -> List[int]:
        # todo: consolidate all ikwargs with same keys to single insert command
        model_ids = []
        for v in ikwargs:
            d = {k: v for k, v in v.items() if model_cls.id_key() not in k}
            keys = list(d.keys())
            values = list(d.values())
            c.execute(
                f'INSERT INTO {model_cls.table_key()} ({", ".join(keys)}) VALUES ({", ".join([self.format_symbol] * len(keys))}) RETURNING {model_cls.id_key()}',
                values,
            )
            model_id = c.fetchone()[0]
            model_ids.append(model_id)

        return model_ids

    @transaction_decorator()
    def _update(self, c, model_cls: IModel, model_id, **ikwargs):
        if len(ikwargs) == 0:
            return

        insert = ", ".join([f"{k} = {self.format_symbol}" for k in ikwargs.keys()])
        values = list(ikwargs.values())
        c.execute(f"UPDATE {model_cls.table_key()} SET {insert} WHERE {model_cls.id_key()} = {model_id};", values)

    @transaction_decorator()
    def update_all(self, c, model_cls: IModel, where: str = None, **ikwargs):
        if len(ikwargs) == 0:
            return

        insert = ", ".join([f"{k} = {self.format_symbol}" for k in ikwargs.keys()])
        values = list(ikwargs.values())

        query_str = f"UPDATE {model_cls.table_key()} SET {insert}"

        if where:
            query_str += f" WHERE {where}"

        c.execute(query_str, values)

    @abstractmethod
    def delete(self, model_cls: IModel, model_id: int):
        pass

    @transaction_decorator()
    def delete_all(self, c, model_cls: IModel, **kwargs):
        query_kwargs = get_query_kwargs(kwargs)
        query_str = f"DELETE FROM {model_cls.table_key()}"

        if "_where" in query_kwargs:
            query_str += f" WHERE {query_kwargs['_where']}"

        c.execute(query_str)

    @transaction_decorator()
    def delete(self, c, model_cls: IModel, model_id: int):
        c.execute(f"DELETE FROM {model_cls.table_key()} WHERE {model_cls.id_key()} = {model_id}")

    @transaction_decorator(exclusive=True)
    def init_db(self, c):
        from ..models import EStatus

        # Create schema version table if not exists
        c.execute("CREATE TABLE IF NOT EXISTS schema_version (" "version INTEGER PRIMARY KEY" ")")
        c.execute("SELECT * FROM schema_version")
        current_schema_version = c.fetchone()
        if current_schema_version is None:
            c.execute(f"INSERT INTO schema_version (version) VALUES ({__schema_version__})")
        else:
            current_schema_version = current_schema_version[0]
            assert (
                current_schema_version == __schema_version__
            ), f"Schema version mismatch, current schema version is {current_schema_version} while code schema version is {__schema_version__}"

        # Create jobs table if not exists
        c.execute(
            "CREATE TABLE IF NOT EXISTS jobs ("
            f"job_id {self.primary_key}, "
            "name TEXT, "
            "description TEXT, "
            "priority REAL DEFAULT 0"
            #   "summary_cookie_keys JSON"
            ")"
        )

        # Create state arguments table if not exists
        c.execute(
            "CREATE TABLE IF NOT EXISTS state_kwargs ("
            f"state_kwargs_id {self.primary_key}, "
            "name TEXT, "
            "entrypoint TEXT NOT NULL, "
            f"targs {self.bytes_type}, "
            "description TEXT, "
            "job_id INTEGER NOT NULL, "
            "CONSTRAINT uq_name_job_id UNIQUE(name, job_id), "
            "CONSTRAINT fk_job_id FOREIGN KEY (job_id) REFERENCES jobs(job_id) ON DELETE CASCADE"
            ")"
        )

        # Create tasks table if not exists
        statuses = ", ".join([f'"{a}"' for a in EStatus])
        c.execute(
            f"CREATE TABLE IF NOT EXISTS tasks ("
            f"task_id {self.primary_key}, "
            "name TEXT, "
            "level REAL, "
            "entrypoint TEXT NOT NULL, "
            f"targs {self.bytes_type}, "
            f"status TEXT ,"  # CHECK(status in ({statuses})),
            f"take_time {self.timestamp_type}, "
            f"start_time {self.timestamp_type}, "
            f"done_time {self.timestamp_type}, "
            f"pulse_time {self.timestamp_type}, "
            "description TEXT, "
            #   "summary_cookie JSON, "
            "job_id INTEGER NOT NULL, "
            "CONSTRAINT fk_job_id FOREIGN KEY (job_id) REFERENCES jobs(job_id) ON DELETE CASCADE"
            ")"
        )

    @transaction_decorator()
    def count_query(
        self, c, model_cls: IModel, _where: str = None, _limit: int = ATASKQ_DB_LIMIT_DEFAULT, _offset: int = 0
    ):
        query_str = f"SELECT COUNT(*) FROM {model_cls.table_key()}"
        query_str = expand_query_str(query_str, _where=_where, _limit=_limit, _offset=_offset)

        c.execute(query_str)

        row = c.fetchone()
        return row[0]

    @transaction_decorator()
    def select_query(
        self,
        c,
        model_cls: IModel,
        _where: str = None,
        _order_by=None,
        _limit: int = ATASKQ_DB_LIMIT_DEFAULT,
        _offset: int = 0,
    ):
        query_str = f"SELECT * FROM {model_cls.table_key()}"
        if _order_by is None:
            _order_by = f"{model_cls.table_key()}.{model_cls.id_key()} ASC"
        query_str = expand_query_str(query_str, _where=_where, _order_by=_order_by, _limit=_limit, _offset=_offset)

        c.execute(query_str)
        rows = c.fetchall()
        col_names = [description[0] for description in c.description]

        return rows, col_names, query_str

    ##################
    # Custom Queries #
    ##################

    @transaction_decorator(exclusive=True)
    def take_next_task(self, c, job_id, level_start, level_stop):
        # imported here to avoid circular dependency
        from ..models import Task, EStatus
        from .handler import Handler, EAction

        # todo: add FOR UPDATE in the queries postgresql
        level_query = ""
        if level_start:
            level_query += f" AND level >= {level_start}"
        if level_stop:
            level_query += f" AND level < {level_stop}"

        # get pending task with minimum level
        query = (
            f"SELECT * FROM tasks WHERE job_id = {job_id} AND status IN ('{EStatus.PENDING}'){level_query} AND level = "
            f"(SELECT MIN(level) FROM tasks WHERE job_id = {job_id} AND status IN ('{EStatus.PENDING}'){level_query})"
            f" {self.for_update}"
        )
        query = query.strip()

        c.execute(query)
        row = c.fetchone()
        if row is None:
            ptask = None
        else:
            col_names = [description[0] for description in c.description]
            ptask = self.from_interface(Task, dict(zip(col_names, row)))

        # get running task with minimum level
        query = (
            f"SELECT * FROM tasks WHERE job_id = {job_id} AND status IN ('{EStatus.RUNNING}'){level_query} AND level = "
            f"(SELECT MIN(level) FROM tasks WHERE job_id = {job_id} AND status IN ('{EStatus.RUNNING}'){level_query})"
            f" {self.for_update}"
        )
        query = query.strip()
        c.execute(query)
        row = c.fetchone()
        if row is None:
            rtask = None
        else:
            col_names = [description[0] for description in c.description]
            rtask = self.from_interface(Task, dict(zip(col_names, row)))

        action = None
        if ptask is None and rtask is None:
            # no more pending task, no more running tasks
            action = EAction.STOP
        elif ptask is None and rtask is not None:
            # no more pending tasks, tasks still running
            action = EAction.WAIT
        elif ptask is not None and rtask is None:
            # pending task next, no more running tasks
            action = EAction.RUN_TASK
        elif ptask is not None and rtask is not None:
            if ptask.level > rtask.level:
                # pending task with level higher than running (wait for running to end)
                action = EAction.WAIT
            elif rtask.level > ptask.level:
                # should never happend
                # running task with level higher than pending task (warn and take next task)
                self.warning(
                    f"Running task with level higher than pending detected, taking pending. running id: {rtask.task_id}, pending id: {ptask.task_id}."
                )
                action = EAction.RUN_TASK
            else:
                action = EAction.RUN_TASK

        if action == EAction.RUN_TASK:
            now = datetime.now()
            c.execute(
                f"UPDATE tasks SET status = '{EStatus.RUNNING}', take_time = {self.timestamp(now)}, pulse_time = {self.timestamp(now)} WHERE task_id = {ptask.task_id};"
            )
            ptask.status = EStatus.RUNNING
            ptask.take_time = now
            ptask.pulse_time = now
            task = ptask
        elif action == EAction.WAIT:
            task = None
        elif action == EAction.STOP:
            task = None
        else:
            raise RuntimeError(f"Unsupported action '{EAction}'")

        return action, task

    @transaction_decorator()
    def tasks_status(
        self,
        c,
        job_id,
        _where: str = None,
        _order_by: str = None,
        _limit: int = ATASKQ_DB_LIMIT_DEFAULT,
        _offset: int = 0,
    ):
        from ..models import EStatus

        query_str = (
            "SELECT level, name,"
            "COUNT(*) as total, "
            + ",".join([f"SUM(CASE WHEN status = '{status}' THEN 1 ELSE 0 END) AS {status} " for status in EStatus])
            + f"FROM tasks WHERE job_id = {job_id} "
            "GROUP BY level, name"
        )

        if _order_by is None:
            _order_by = "name ASC"

        query_str = expand_query_str(query_str, _order_by=_order_by, _limit=_limit, _offset=_offset)

        c.execute(query_str)
        rows = c.fetchall()
        col_names = [description[0] for description in c.description]

        ret = [dict(zip(col_names, row)) for row in rows]
        return ret

    @transaction_decorator()
    def jobs_status(
        self, c, _where: str = None, _order_by: str = None, _limit: int = ATASKQ_DB_LIMIT_DEFAULT, _offset: int = 0
    ):
        from ..models import EStatus

        query_str = (
            "SELECT jobs.job_id, jobs.name, jobs.description, jobs.priority, "
            "COUNT(*) as tasks, "
            + ", ".join([f"SUM(CASE WHEN status = '{status}' THEN 1 ELSE 0 END) AS {status}" for status in EStatus])
            + f" FROM jobs "
            "LEFT JOIN tasks ON jobs.job_id = tasks.job_id "
            "GROUP BY jobs.job_id"
        )

        if _order_by is None:
            _order_by = "jobs.job_id DESC"

        query_str = expand_query_str(query_str, _order_by=_order_by, _limit=_limit, _offset=_offset)

        c.execute(query_str)
        rows = c.fetchall()
        col_names = [description[0] for description in c.description]

        ret = [dict(zip(col_names, row)) for row in rows]
        return ret

    @transaction_decorator()
    def fail_pulse_timeout_tasks(self, c, timeout_sec=None):
        from ..models import EStatus

        if timeout_sec is None:
            return

        # set timeout tasks
        last_valid_pulse = datetime.now() - timedelta(seconds=timeout_sec)
        c.execute(
            f"UPDATE tasks SET status = '{EStatus.FAILURE}' WHERE pulse_time < {self.timestamp(last_valid_pulse)} AND status NOT IN ('{EStatus.SUCCESS}', '{EStatus.FAILURE}');"
        )



from multiprocessing import Lock


def write_to_file(filepath, text):
    with open(filepath, "a") as f:
        f.write(text)


__lock__ = Lock()


def write_to_file_mp_lock(filepath, text):
    with __lock__:
        with open(filepath, "a") as f:
            f.write(text)



import logging


class Logger:
    def __init__(self, logger: logging.Logger or None):
        self._logger = logger or logging.getLogger("ataskq")

    def exception(self, *args, **kwargs):
        self._logger.exception(*args, **kwargs)

    def critical(self, *args, **kwargs):
        self._logger.critical(*args, **kwargs)

    def error(self, *args, **kwargs):
        self._logger.error(*args, **kwargs)

    def warning(self, *args, **kwargs):
        self._logger.warning(*args, **kwargs)

    def info(self, *args, **kwargs):
        self._logger.info(*args, **kwargs)

    def debug(self, *args, **kwargs):
        self._logger.debug(*args, **kwargs)
